import {ShaderPass} from "../framebuffers/ShaderPass.js";

/*** FXAAPASS CLASS ***/

/***
 Here we create our FXAAPass object
 This is just a regular ShaderPass with preset shaders and a resolution uniform

 params: see ShaderPas class object

 returns :
 @this: our FXAAPass element
 ***/
export class FXAAPass extends ShaderPass {
    constructor(curtains, {
        // Mesh params
        renderOrder,
        depthTest,
        texturesOptions,
        crossOrigin,

        // ShaderPass specific params
        depth,
        clear,
        renderTarget,
    } = {}) {

        // taken from https://github.com/spite/Wagner/blob/master/fragment-shaders/fxaa-fs.glsl
        const fragmentShader = `
            precision mediump float;
            
            varying vec3 vVertexPosition;
            varying vec2 vTextureCoord;
        
            uniform sampler2D uRenderTexture;
            
            uniform vec2 uResolution;
            
            #define FXAA_REDUCE_MIN   (1.0/128.0)
            #define FXAA_REDUCE_MUL   (1.0/8.0)
            #define FXAA_SPAN_MAX     8.0
            
            void main() {
                vec2 res = 1.0 / uResolution;
            
                vec3 rgbNW = texture2D(uRenderTexture, (vTextureCoord.xy + vec2(-1.0, -1.0) * res)).xyz;
                vec3 rgbNE = texture2D(uRenderTexture, (vTextureCoord.xy + vec2(1.0, -1.0) * res)).xyz;
                vec3 rgbSW = texture2D(uRenderTexture, (vTextureCoord.xy + vec2(-1.0, 1.0) * res)).xyz;
                vec3 rgbSE = texture2D(uRenderTexture, (vTextureCoord.xy + vec2(1.0, 1.0) * res)).xyz;
                vec4 rgbaM = texture2D(uRenderTexture, vTextureCoord.xy * res);
                vec3 rgbM = rgbaM.xyz;
                vec3 luma = vec3(0.299, 0.587, 0.114);
            
                float lumaNW = dot(rgbNW, luma);
                float lumaNE = dot(rgbNE, luma);
                float lumaSW = dot(rgbSW, luma);
                float lumaSE = dot(rgbSE, luma);
                float lumaM  = dot(rgbM,  luma);
                float lumaMin = min(lumaM, min(min(lumaNW, lumaNE), min(lumaSW, lumaSE)));
                float lumaMax = max(lumaM, max(max(lumaNW, lumaNE) , max(lumaSW, lumaSE)));
            
                vec2 dir;
                dir.x = -((lumaNW + lumaNE) - (lumaSW + lumaSE));
                dir.y = ((lumaNW + lumaSW) - (lumaNE + lumaSE));
            
                float dirReduce = max((lumaNW + lumaNE + lumaSW + lumaSE) * (0.25 * FXAA_REDUCE_MUL), FXAA_REDUCE_MIN);
            
                float rcpDirMin = 1.0 / (min(abs(dir.x), abs(dir.y)) + dirReduce);
                dir = min(vec2(FXAA_SPAN_MAX, FXAA_SPAN_MAX),
                      max(vec2(-FXAA_SPAN_MAX, -FXAA_SPAN_MAX),
                            dir * rcpDirMin)) * res;
                vec4 rgbA = (1.0/2.0) * (
                texture2D(uRenderTexture, vTextureCoord.xy + dir * (1.0/3.0 - 0.5)) +
                texture2D(uRenderTexture, vTextureCoord.xy + dir * (2.0/3.0 - 0.5)));
                vec4 rgbB = rgbA * (1.0/2.0) + (1.0/4.0) * (
                texture2D(uRenderTexture, vTextureCoord.xy + dir * (0.0/3.0 - 0.5)) +
                texture2D(uRenderTexture, vTextureCoord.xy + dir * (3.0/3.0 - 0.5)));
                float lumaB = dot(rgbB, vec4(luma, 0.0));
            
                if ((lumaB < lumaMin) || (lumaB > lumaMax)) {
                    gl_FragColor = rgbA;
                } else {
                    gl_FragColor = rgbB;
                }
            }
        `;

        const uniforms = {
            resolution: {
                name: "uResolution",
                type: "2f",
                value: [0, 0], // will be updated after having called super()
            }
        };

        super(curtains, {
            fragmentShader,
            uniforms,

            // Mesh params
            renderOrder,
            depthTest,
            texturesOptions,
            crossOrigin,
            depth,
            clear,
            renderTarget,
        });

        // return if no gl context
        if(!this.gl) {
            return;
        }

        // update the resolution uniform
        this.uniforms.resolution.value = [
            this.renderer._boundingRect.width,
            this.renderer._boundingRect.height
        ];

        // override onAfterResize callback
        this._onAfterResizeCallback = () => {
            // update the resolution uniform
            this.uniforms.resolution.value = [
                this.renderer._boundingRect.width,
                this.renderer._boundingRect.height
            ];

            this._onFXAAPassAfterResizeCallback && this._onFXAAPassAfterResizeCallback();
        };
    }

    /***
     This is called each time the FXAAPass has been resized

     params :
     @callback (function) : a function to execute

     returns :
     @this: our FXAAPass to handle chaining
     ***/
    onAfterResize(callback) {
        if(callback) {
            this._onFXAAPassAfterResizeCallback = callback;
        }

        return this;
    }
}